unit RessourceAction;

interface

uses Windows, SysUtils, Classes, Graphics, JPEG, GIFImage,
     unitPEFile, unitResourceDetails, unitResourceGraphics,
     unitResourceXPManifests, unitResourceGIF,
     unitResourceJPEG, unitResourceRCData;

type TRessources = record
     Classe:TResourceDetailsClass;
     Types: string;
     Extensions: string
end;

const
  ClassesRessources : array [0..6] of TRessources = (
    (Classe: TCursorGroupResourceDetails ; Types: 'Curseurs'; Extensions : '*.cur'),
    (Classe: TBitmapResourceDetails ; Types: 'Images bitmap'; Extensions : '*.bmp'),
    (Classe: TIconGroupResourceDetails ; Types: 'Icnes'; Extensions : '*.ico'),
    (Classe: TJPegResourceDetails ; Types: 'Images JPG'; Extensions : '*.jpg;*.jpeg'),
    (Classe: TGIFResourceDetails ; Types: 'Images GIF'; Extensions : '*.gif'),
    (Classe: TXPManifestResourceDetails ; Types: 'Fichiers Manifest'; Extensions : '*.manifest'),
    (Classe: TRCDataPackagesResourceDetails ; Types: 'Datas'; Extensions : '*.*')
    );

   RES_AJOUT=0;
   RES_MODIFICATION=1;
   RES_SUPPRESSION=2;

function ChargerRessource(FileName: TFileName): bool;
procedure ListerRessources(TypeRessource: integer ; Liste: TStrings);
function ModifierRessource(TypeModification: integer ; TypeRessource: integer ; NomRessource: string ; const FileName: TFileName='' ; const Langue: integer=0): bool;
function SauverRessource(FileName: TFileName): bool;
procedure DechargerRessource;

var PEModule: TPEResourceModule;

implementation

function ChargerRessource(FileName: TFileName): bool;
begin
     result:=true;
     try
        if Assigned(PEModule) then DechargerRessource;
        PEModule:=TPEResourceModule.Create;
        PEModule.LoadFromFile(FileName);
     except result:=false;
     end;
end;

procedure ListerRessources(TypeRessource: integer ; Liste: TStrings);
var i: integer;
begin
     if not Assigned(PEModule) then exit;
     Liste.Clear;
     for i:=0 to PEModule.ResourceCount-1 do
        if PEModule.ResourceDetails[i] is ClassesRessources[TypeRessource].Classe then
           Liste.Add(PEModule.ResourceDetails[i].ResourceName);
end;


function FindResource{NoLanguage}(const tp, Name: string): TResourceDetails;
var
  i : Integer;
begin
  Result := nil;
  for i := 0 to PEModule.ResourceCount - 1 do
    if (PEModule.ResourceDetails [i].ResourceType = tp) and SameText(PEModule.ResourceDetails [i].ResourceName,Name) then
    begin
      Result := PEModule.ResourceDetails [i];
      break
    end;
end;


function ModifierRessource(TypeModification: integer ; TypeRessource: integer ; NomRessource: string ; const FileName: TFileName='' ; const Langue: integer=0): bool;
var Ressource: TResourceDetails;
    Picture: TPicture;
    MemoryStream: TMemoryStream;
    Lignes: TStringList;
begin
     result:=false;
     if not Assigned(PEModule) then exit;
     try
        case TypeModification of
        RES_AJOUT:
        begin
             case TypeRessource of
             0:
             begin
                  Ressource:=(ClassesRessources[TypeRessource].Classe).CreateNew(PEModule, Langue, NomRessource);
                  TCursorGroupResourceDetails(Ressource).LoadImage(FileName);
                  result:=true;
             end;
             1:
             begin
                  Ressource:=(ClassesRessources[TypeRessource].Classe).CreateNew(PEModule, Langue, NomRessource);
                  Picture:=TPicture.Create;
                  Picture.LoadFromFile(FileName);
                  TBitmapResourceDetails(Ressource).SetImage(Picture);
                  Picture.Free;
                  result:=true;
             end;
             2:
             begin
                  Ressource:=(ClassesRessources[TypeRessource].Classe).CreateNew(PEModule, Langue, NomRessource);
                  TIconGroupResourceDetails(Ressource).LoadImage(FileName);
                  result:=true;
             end;
             3:
             begin
                  Ressource:=(ClassesRessources[TypeRessource].Classe).CreateNew(PEModule, Langue, NomRessource);
                  Picture:=TPicture.Create;
                  Picture.LoadFromFile(FileName);
                  TJPegResourceDetails(Ressource).SetImage(Picture);
                  Picture.Free;
                  result:=true;
             end;
             4:
             begin
                  Ressource:=(ClassesRessources[TypeRessource].Classe).CreateNew(PEModule, Langue, NomRessource);
                  Picture:=TPicture.Create;
                  Picture.LoadFromFile(FileName);
                  TGIFResourceDetails(Ressource).SetImage(Picture);
                  Picture.Free;
                  result:=true;
             end;
             5:
             begin
                  Ressource:=(ClassesRessources[TypeRessource].Classe).CreateNew(PEModule, Langue, NomRessource);
                  Lignes:=TStringList.Create;
                  Lignes.LoadFromFile(FileName);
                  TXPManifestResourceDetails(Ressource).SetText(Lignes);
                  Lignes.Free;
                  result:=true;
             end;
             6:
             begin
                  MemoryStream:=TMemoryStream.Create;
                  MemoryStream.LoadFromFile(FileName);
                  Ressource := TResourceDetails.CreateResourceDetails (PEModule, Langue, UpperCase(NomRessource),
                                                                       IntToStr (Integer (RT_RCDATA)), MemoryStream.Size, MemoryStream.Memory);
                  PEModule.AddResource (Ressource);
                  MemoryStream.Free;
                  result:=true;
             end;
             end;
        end;
        RES_MODIFICATION:
        begin
             result:=ModifierRessource(RES_SUPPRESSION, TypeRessource, NomRessource)
                 and ModifierRessource(RES_AJOUT, TypeRessource, NomRessource, FileName);
        end;
        RES_SUPPRESSION:
        begin
             Ressource:=FindResource(ClassesRessources[TypeRessource].Classe.GetBaseType, NomRessource);
             if Assigned(Ressource) then
             begin
                  PEModule.DeleteResource(PEModule.IndexOfResource(Ressource));
                  result:=true
             end;
        end;
        end
     except
     end;
     PEModule.Dirty:=true;
     PEModule.SortResources;
end;

function SauverRessource(FileName: TFileName): bool;
begin
     result:=true; if FileExists(FileName) then result:=DeleteFile(FileName);

     if result then
        if Assigned(PEModule) then
        try
           PEModule.SaveToFile(FileName);
        except
           result:=false;
        end
        else result:=false;
end;

procedure DechargerRessource;
begin
     if Assigned(PEModule) then PEModule.Free;
     PEModule:=nil;
end;

end.
