{
  L'unit RessourceAction regroupe les procdures utiles pour
  ajouter/modifier/supprimer facilement des ressources d'un fichier (d'extension
  EXE, DLL, CPL, OCX, BPL, SCR), cr avec Delphi.

  Appeler successivement :

  1. function ChargerRessource(FileName: TFileName): bool;
    > Pour charger les ressources de FileName en mmoire.

  2. function ModifierRessource(TypeModification: integer ; TypeRessource: integer;
                                NomRessource: string ; const FileName: TFileName='';
                                const Langue: integer=0): bool;
    > Pour effectuer la modifications sur le module charg en mmoire :

       TypeModification vaut :
          RES_AJOUT pour ajouter une ressource
          RES_MODIFICATION pour modifier une ressource
          RES_SUPPRESSION pour supprimer une ressource;

       TypeRessource vaut :
          0 pour les curseurs (RT_CURSOR)
          1 pour les images bitmaps  (RT_BITMAP)
          2 pour les icnes (RT_ICON)
          3 pour les images jpeg
          4 pour les images gif
          5 pour les fichiers XP manifest
          6 pour les donnes (RT_DATAS)
          (cette liste peut tre complte);

       NomRessource est le nom de la ressource  modifier;

       FileName est le fichier de la ressource
       (uniquement pour RES_AJOUT et RES_MODIFICATION);

       Langue est le code de langue (0 par dfaut).

  Important :
  -----------
    Cette tape peut tre rpte autant de fois que ncessaire pour modifier plusieurs
    ressouces dans le fichier charg mmoire.

  3. function SauverRessource(FileName: TFileName): bool;
    > Pour enregistrer le fichier contenant les modifications sur le disque dur.

  4. procedure DechargerRessource;
    > Pour librer la mmoire.


  Facultatif :
  ------------
  procedure ListerRessources(TypeRessource: integer ; Liste: TStrings);
    > Cette procedure renvoie dans Liste la liste des ressources
      du type TypeRessource (voir plus haut pour les types) du module
      charg en mmoire
}

unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls;
                                  
type
  TForm1 = class(TForm)
    Source: TEdit;
    LabelSource: TLabel;
    LabelDestination: TLabel;
    Destination: TEdit;
    SelectionSource: TButton;
    SelectionDestination: TButton;
    Ouvrir: TOpenDialog;
    Enregistrer: TSaveDialog;
    Appliquer: TButton;
    Label8: TLabel;
    Panel1: TPanel;
    Operation: TComboBox;
    TypeRessource: TComboBox;
    LabelOperation: TLabel;
    LabelType: TLabel;
    ListeRessources: TListBox;
    LabelListe: TLabel;
    NomRessource: TEdit;
    LabelRessource: TLabel;
    LabelFichier: TLabel;
    FichierRessource: TEdit;
    SelectionRessource: TButton;
    OuvrirRessource: TOpenDialog;
    Charger: TButton;
    Decharger: TButton;
    procedure SelectionSourceClick(Sender: TObject);
    procedure SelectionDestinationClick(Sender: TObject);
    procedure SelectionRessourceClick(Sender: TObject);
    procedure DechargerClick(Sender: TObject);
    procedure SourceChange(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ChargerClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure OperationChange(Sender: TObject);
    procedure FichierRessourceChange(Sender: TObject);
    procedure AppliquerClick(Sender: TObject);
    procedure TypeRessourceChange(Sender: TObject);
    procedure ListeRessourcesClick(Sender: TObject);
    procedure DestinationChange(Sender: TObject);
  private
    { Dclarations prives }
  public
    { Dclarations publiques }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

uses RessourceAction;

procedure TForm1.SelectionSourceClick(Sender: TObject);
begin if Ouvrir.Execute then Source.Text:=Ouvrir.FileName; end;

procedure TForm1.SelectionDestinationClick(Sender: TObject);
begin
     Enregistrer.FileName:=ExtractFileName(Source.Text);
     if Enregistrer.Execute then Destination.Text:=Enregistrer.FileName;
end;

procedure TForm1.SelectionRessourceClick(Sender: TObject);
begin
     OuvrirRessource.Filter:=ClassesRessources[TypeRessource.ItemIndex].Types+
                             ' ('+ClassesRessources[TypeRessource.ItemIndex].Extensions+')|'+
                             ClassesRessources[TypeRessource.ItemIndex].Extensions;
     if OuvrirRessource.Execute then FichierRessource.Text:=OuvrirRessource.FileName;
end;

procedure TForm1.DechargerClick(Sender: TObject);
var i: integer;
begin
     DechargerRessource;

     LabelSource.Enabled:=true;
     Source.Enabled:=true;
     SelectionSource.Enabled:=true;
     Charger.Enabled:=true;

     Decharger.Enabled:=false;
     for i:=0 to Panel1.ControlCount-1 do Panel1.Controls[i].Enabled:=false;
     Appliquer.Enabled:=false;
     Destination.Enabled:=false;
     SelectionDestination.Enabled:=false;
     LabelDestination.Enabled:=false;
end;

procedure TForm1.SourceChange(Sender: TObject);
begin Charger.Enabled:=not Decharger.Enabled and FileExists(Source.Text); end;

procedure TForm1.FormCreate(Sender: TObject);
var i: integer;
begin
     Source.Text:=ParamStr(0);
     for i:=0 to Pred(length(ClassesRessources)) do
        TypeRessource.Items.Add(ClassesRessources[i].Types);
     TypeRessource.ItemIndex:=0;
end;

procedure TForm1.ChargerClick(Sender: TObject);
var i: integer;
begin
     if ChargerRessource(Source.Text) then
     begin
          Tag:=0;
          SelectionSource.Enabled:=false;
          LabelSource.Enabled:=false;
          Source.Enabled:=false;
          Charger.Enabled:=false;

          Decharger.Enabled:=true;

          for i:=0 to Panel1.ControlCount-1 do Panel1.Controls[i].Enabled:=true;
          OperationChange(Sender);
          Operation.Setfocus;
     end else ShowMessage('Fichier incompatible !');
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin DechargerRessource; end;

procedure TForm1.OperationChange(Sender: TObject);
begin
     LabelFichier.Enabled:=not (Operation.ItemIndex=RES_SUPPRESSION);
     FichierRessource.Enabled:=LabelFichier.Enabled;
     SelectionRessource.Enabled:=LabelFichier.Enabled;
     TypeRessourceChange(Sender);
end;

procedure TForm1.TypeRessourceChange(Sender: TObject);
begin
     ListerRessources(TypeRessource.ItemIndex, ListeRessources.Items);
     FichierRessourceChange(Sender);
end;

procedure TForm1.FichierRessourceChange(Sender: TObject);
var k: integer;
begin
     k:=NomRessource.SelStart;
     NomRessource.Text:=Trim(UpperCase(NomRessource.Text));
     NomRessource.SelStart:=k;
     case Operation.ItemIndex of
     RES_AJOUT:
        LabelDestination.Enabled:=(NomRessource.Text<>'') and (ListeRessources.Items.IndexOf(NomRessource.Text)=-1) and FileExists(FichierRessource.Text);
     RES_MODIFICATION:
        LabelDestination.Enabled:=(NomRessource.Text<>'') and (ListeRessources.Items.IndexOf(NomRessource.Text)<>-1) and FileExists(FichierRessource.Text);
     RES_SUPPRESSION:
        LabelDestination.Enabled:=(NomRessource.Text<>'') and (ListeRessources.Items.IndexOf(NomRessource.Text)<>-1);
     end;
     Destination.Enabled:=LabelDestination.Enabled;
     SelectionDestination.Enabled:=LabelDestination.Enabled;
     DestinationChange(Sender);
end;

procedure TForm1.AppliquerClick(Sender: TObject);
begin
     Tag:=Tag+1;
     if ModifierRessource(Operation.ItemIndex,TypeRessource.ItemIndex, NomRessource.Text, FichierRessource.Text)
       and SauverRessource(Destination.Text) then
          MessageDlg(#13'Modification n'+inttostr(Tag)+' apporte au module avec succs !',mtInformation, [mbOk],0)
     else
     begin
          MessageDlg(#13'Erreur lors de la modification n'+inttostr(Tag)+'.',mtError, [mbOk],0);
          Tag:=Tag-1;
     end;
     ListerRessources(TypeRessource.ItemIndex, ListeRessources.Items);
end;

procedure TForm1.ListeRessourcesClick(Sender: TObject);
begin if ListeRessources.ItemIndex<>-1 then NomRessource.Text:=ListeRessources.Items[ListeRessources.ItemIndex]; end;



procedure TForm1.DestinationChange(Sender: TObject);
begin
     Appliquer.Enabled:=Destination.Enabled and (Trim(Destination.Text)<>'');
end;

end.
